@@ -1,20 +1,26 @@
+//! Create, start, introspect, stop, and destroy PostgreSQL clusters.
+
+use std::ffi::{OsStr, OsString};
+use std::os::unix::prelude::OsStringExt;
 use std::path::{Path, PathBuf};
 use std::process::{Command, ExitStatus, Output};
 use std::{env, error, fmt, fs, io};
 
-use shell_escape::escape;
+use nix::errno::Errno;
+use shell_quote::sh::escape_into;
 
-use crate::lock::LockDo;
 use crate::runtime;
+use crate::version;
 
 #[derive(Debug)]
 pub enum ClusterError {
     PathEncodingError, // Path is not UTF-8.
     IoError(io::Error),
     UnixError(nix::Error),
-    UnsupportedVersion(runtime::Version),
-    UnknownVersion(runtime::VersionError),
+    UnsupportedVersion(version::Version),
+    UnknownVersion(version::VersionError),
     DatabaseError(postgres::error::Error),
+    InUse, // Cluster is already in use; cannot lock exclusively.
     Other(Output),
 }
 
@@ -28,6 +34,7 @@ impl fmt::Display for ClusterError {
             UnsupportedVersion(ref e) => write!(fmt, "PostgreSQL version not supported: {}", e),
             UnknownVersion(ref e) => write!(fmt, "PostgreSQL version not known: {}", e),
             DatabaseError(ref e) => write!(fmt, "database error: {}", e),
+            InUse => write!(fmt, "cluster in use; cannot lock exclusively"),
             Other(ref e) => write!(fmt, "external command failed: {:?}", e),
         }
     }
@@ -42,6 +49,7 @@ impl error::Error for ClusterError {
             ClusterError::UnsupportedVersion(_) => None,
             ClusterError::UnknownVersion(ref error) => Some(error),
             ClusterError::DatabaseError(ref error) => Some(error),
+            ClusterError::InUse => None,
             ClusterError::Other(_) => None,
         }
     }
@@ -59,8 +67,8 @@ impl From<nix::Error> for ClusterError {
     }
 }
 
-impl From<runtime::VersionError> for ClusterError {
-    fn from(error: runtime::VersionError) -> ClusterError {
+impl From<version::VersionError> for ClusterError {
+    fn from(error: version::VersionError) -> ClusterError {
         ClusterError::UnknownVersion(error)
     }
 }
@@ -71,23 +79,26 @@ impl From<postgres::error::Error> for ClusterError {
     }
 }
 
+/// Representation of a PostgreSQL cluster.
+///
+/// The cluster may not yet exist on disk. It may exist but be stopped, or it
+/// may be running. The methods here can be used to create, start, introspect,
+/// stop, and destroy the cluster. There's no protection against concurrent
+/// changes to the cluster made by other processes, but the functions in the
+/// [`coordinate`][`crate::coordinate`] module may help.
 pub struct Cluster {
     /// The data directory of the cluster.
     ///
     /// Corresponds to the `PGDATA` environment variable.
     datadir: PathBuf,
-    /// Lock file.
-    lockfile: PathBuf,
     /// The installation of PostgreSQL to use with this cluster.
     runtime: runtime::Runtime,
 }
 
 impl Cluster {
     pub fn new<P: AsRef<Path>>(datadir: P, runtime: runtime::Runtime) -> Self {
-        let datadir = datadir.as_ref();
         Self {
-            datadir: datadir.to_path_buf(),
-            lockfile: datadir.parent().unwrap_or(datadir).join(".cluster.lock"),
+            datadir: datadir.as_ref().to_path_buf(),
             runtime,
         }
     }
@@ -99,14 +110,16 @@ impl Cluster {
         command
     }
 
+    /// A fairly simplistic check: does the data directory exist and does it
+    /// contain a file named `PG_VERSION`?
     pub fn exists(&self) -> bool {
         self.datadir.is_dir() && self.datadir.join("PG_VERSION").is_file()
     }
 
     /// Check if this cluster is running.
     ///
-    /// Tries to distinguish carefully between "definitely running", "definitely not running", and
-    /// "don't know". The latter results in `ClusterError`.
+    /// Tries to distinguish carefully between "definitely running", "definitely
+    /// not running", and "don't know". The latter results in `ClusterError`.
     pub fn running(&self) -> Result<bool, ClusterError> {
         let output = self.ctl().arg("status").output()?;
         let code = match output.status.code() {
@@ -213,21 +226,17 @@ impl Cluster {
         self.datadir.join("backend.log")
     }
 
-    /// Return an open `File` for this cluster's lock file.
-    fn lock(&self) -> io::Result<fs::File> {
-        fs::OpenOptions::new()
-            .append(true)
-            .create(true)
-            .open(&self.lockfile)
-    }
-
     /// Create the cluster if it does not already exist.
     pub fn create(&self) -> Result<bool, ClusterError> {
-        self.lock()?.do_exclusive(|| self._create())?
+        match self._create() {
+            Err(ClusterError::UnixError(Errno::EAGAIN)) if self.exists() => Ok(false),
+            Err(ClusterError::UnixError(Errno::EAGAIN)) => Err(ClusterError::InUse),
+            other => other,
+        }
     }
 
     fn _create(&self) -> Result<bool, ClusterError> {
-        match self.datadir.join("PG_VERSION").is_file() {
+        match self.exists() {
             // Nothing more to do; the cluster is already in place.
             true => Ok(false),
             // Create the cluster and report back that we did so.
@@ -246,38 +255,21 @@ impl Cluster {
 
     // Start the cluster if it's not already running.
     pub fn start(&self) -> Result<bool, ClusterError> {
-        self.lock()?.do_exclusive(|| self._start())?
+        match self._start() {
+            Err(ClusterError::UnixError(Errno::EAGAIN)) if self.running()? => Ok(false),
+            Err(ClusterError::UnixError(Errno::EAGAIN)) => Err(ClusterError::InUse),
+            other => other,
+        }
     }
 
     fn _start(&self) -> Result<bool, ClusterError> {
-        // If the cluster's already running, don't do anything.
-        if self.running()? {
-            return Ok(false);
-        }
         // Ensure that the cluster has been created.
         self._create()?;
-        // This next thing is a wart of the `shell-escape` crate. UNIX paths are
-        // bytes and the only thing they're not allowed to contain is, as far as
-        // I know, the null byte. The encoding is defined by the locale.
-        //
-        // Now, the `shell-escape` crate only understands `&str` so we have to
-        // convert a platform-specific path string to a UTF-8 string before
-        // passing it to `shell-escape`.
-        //
-        // Cargo, a notable codebase that uses `shell-escape`, sidesteps this
-        // limitation: it only uses `shell-escape` with strings already assumed
-        // to be UTF-8. I'm choosing to be strict and reject any platform path
-        // that's not also valid UTF-8.
-        //
-        // Why do we need `shell-escape`? One of the arguments we will pass to
-        // `pg_ctl` will be used as the argument _list_ when it invokes
-        // `postgres`. Sucks, but there it is.
-        let datadir = self
-            .datadir
-            .as_path()
-            .as_os_str()
-            .to_str()
-            .ok_or(ClusterError::PathEncodingError)?;
+        // Check if we're running already.
+        if self.running()? {
+            // We didn't start this cluster; say so.
+            return Ok(false);
+        }
         // Next, invoke `pg_ctl` to start the cluster.
         // pg_ctl options:
         //  -l <file> -- log file.
@@ -294,7 +286,11 @@ impl Cluster {
             .arg("-s")
             .arg("-w")
             .arg("-o")
-            .arg(format!("-h '' -F -k {}", escape(datadir.into())))
+            .arg({
+                let mut arg = b"-h '' -F -k "[..].into();
+                escape_into(&self.datadir, &mut arg);
+                OsString::from_vec(arg)
+            })
             .output()?;
         // We did actually start the cluster; say so.
         Ok(true)
@@ -314,16 +310,34 @@ impl Cluster {
 
     pub fn shell(&self, database: &str) -> Result<ExitStatus, ClusterError> {
         let mut command = self.runtime.execute("psql");
-        command.arg("--quiet").arg("--").arg(database);
+        command.arg("--quiet");
         command.env("PGDATA", &self.datadir);
         command.env("PGHOST", &self.datadir);
+        command.env("PGDATABASE", database);
+        Ok(command.spawn()?.wait()?)
+    }
+
+    pub fn exec<T: AsRef<OsStr>>(
+        &self,
+        database: &str,
+        command: T,
+        args: &[T],
+    ) -> Result<ExitStatus, ClusterError> {
+        let mut command = self.runtime.command(command);
+        command.args(args);
+        command.env("PGDATA", &self.datadir);
+        command.env("PGHOST", &self.datadir);
+        command.env("PGDATABASE", database);
         Ok(command.spawn()?.wait()?)
     }
 
     // The names of databases in this cluster.
     pub fn databases(&self) -> Result<Vec<String>, ClusterError> {
         let mut conn = self.connect("template1")?;
-        let rows = conn.query("SELECT datname FROM pg_catalog.pg_database", &[])?;
+        let rows = conn.query(
+            "SELECT datname FROM pg_catalog.pg_database ORDER BY datname",
+            &[],
+        )?;
         let datnames: Vec<String> = rows.iter().map(|row| row.get(0)).collect();
         Ok(datnames)
     }
@@ -346,7 +360,11 @@ impl Cluster {
 
     // Stop the cluster if it's running.
     pub fn stop(&self) -> Result<bool, ClusterError> {
-        self.lock()?.do_exclusive(|| self._stop())?
+        match self._stop() {
+            Err(ClusterError::UnixError(Errno::EAGAIN)) if !self.running()? => Ok(false),
+            Err(ClusterError::UnixError(Errno::EAGAIN)) => Err(ClusterError::InUse),
+            other => other,
+        }
     }
 
     fn _stop(&self) -> Result<bool, ClusterError> {
@@ -369,7 +387,10 @@ impl Cluster {
 
     // Destroy the cluster if it exists, after stopping it.
     pub fn destroy(&self) -> Result<bool, ClusterError> {
-        self.lock()?.do_exclusive(|| self._destroy())?
+        match self._destroy() {
+            Err(ClusterError::UnixError(Errno::EAGAIN)) => Err(ClusterError::InUse),
+            other => other,
+        }
     }
 
     fn _destroy(&self) -> Result<bool, ClusterError> {
@@ -397,7 +418,7 @@ mod tests {
             println!("{:?}", runtime);
             let cluster = Cluster::new("some/path", runtime);
             assert_eq!(Path::new("some/path"), cluster.datadir);
-            assert_eq!(false, cluster.running().unwrap());
+            assert!(!cluster.running().unwrap());
         }
     }
 
